import 'dart:async';

enum LifecycleState {
  created,
  resumed,
  inactive,
  paused,
  detached,
  stopped,
  destroyed,
}

extension LifecycleStateUtils on LifecycleState {
  static const ALL_STATES = {
    LifecycleState.created,
    LifecycleState.resumed,
    LifecycleState.inactive,
    LifecycleState.paused,
    LifecycleState.detached,
    LifecycleState.stopped,
    LifecycleState.destroyed,
  };
}

typedef RecreateOnResume<T> = StreamSubscription<T> Function();

class StreamSubscriptionLifecycleAware {
  StreamSubscription? streamSubscription;
  final Set<LifecycleState>? lifecycleStates;
  final RecreateOnResume? recreateOnResume;

  StreamSubscriptionLifecycleAware(
      {this.streamSubscription, this.lifecycleStates, this.recreateOnResume});

  void onLifecycleChanged(LifecycleState lifecycleState) {
    switch (lifecycleState) {
      case LifecycleState.resumed:
        resume(LifecycleState.resumed);
        break;
      case LifecycleState.paused:
        pause(LifecycleState.paused);
        break;
      case LifecycleState.stopped:
        pause(LifecycleState.stopped);
        break;
      case LifecycleState.destroyed:
        cancel(LifecycleState.destroyed);
        break;
      default:
    }
  }

  bool allowed(LifecycleState lifecycleState) =>
      lifecycleStates?.contains(lifecycleState) ?? false;

  void resume(LifecycleState lifecycleState) {
    if (recreateOnResume != null) {
      streamSubscription = recreateOnResume!.call();
    } else if (allowed(lifecycleState)) {
      streamSubscription?.resume();
    }
  }

  void pause(LifecycleState lifecycleState, [Future? resumeSignal]) {
    if (recreateOnResume != null) {
      cancel(LifecycleState.destroyed);
    } else if (allowed(lifecycleState)) {
      streamSubscription?.pause(resumeSignal);
    }
  }

  void cancel(LifecycleState lifecycleState) {
    if (allowed(lifecycleState)) streamSubscription?.cancel();
  }
}
