import 'package:adeptutils/src/wrapper/modal_progress_hud.dart';
import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/collections/stack.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:adeptutils/src/exception_utils.dart';
import 'package:adeptutils/src/enumerators/lifecycle_state.dart';

/// This allows a value of type T or T? to be treated as a value of type T?.
///
/// We use this so that APIs that have become non-nullable can still be used
/// with `!` and `?` on the stable branch.
T? _ambiguate<T>(T? value) => value;

abstract class AdeptState<W extends StatefulWidget, T, B> extends State<W>
    with WidgetsBindingObserver
    implements IWidget<T> {
  final _scaffoldKey = GlobalKey<ScaffoldState>();
  GlobalKey<ScaffoldState> get scaffoldKey => _scaffoldKey;
  final B? _base;
  B? get base => _base;
  int _exitControl = 0;
  String? _loadingMessage;
  Widget? _progressIndicator;
  bool _isLoading = false,
      _showProgressIndeterminate = true,
      _pressBackAgain = false,
      _isBuildCompleted = false;
  get isBuildCompleted => _isBuildCompleted;
  String _pressBackAgainMessage;
  Widget fragment = const SizedBox();
  ListStack<Widget> fragmentsManager = ListStack();
  LifecycleState _lifecycleState = LifecycleState.created;
  LifecycleState get lifecycleState => _lifecycleState;
  final Set<StreamSubscriptionLifecycleAware>
      _streamSubscriptionsLifecycleAware = {};

  AdeptState(
      {B? base, bool pressBackAgain = false, String? pressBackAgainMessage})
      : _base = base,
        _pressBackAgain = pressBackAgain,
        _pressBackAgainMessage =
            pressBackAgainMessage ?? 'Press back again to exit';

  void handleException(dynamic e,
      {BuildContext? context,
      String title = 'Error',
      String btnContinueText = 'CONTINUE',
      String btnExitText = 'EXIT'}) {
    context ??= this.context;
    ExceptionUtils.handleException(e, context,
        title: title,
        btnContinueText: btnContinueText,
        btnExitText: btnExitText);
  }

  ///The streamSubscription to be aware of life cycle changes
  ///The lifecycleStates to take into account during changes, by default all lifeCycles will be taken into account
  ///The recreateOnResume function indicates that the streamSubscription should cancel on pause and recreate on resume by using this function
  bool addLifecycleAware(StreamSubscription? streamSubscription,
          {Set<LifecycleState> lifecycleStates = LifecycleStateUtils.ALL_STATES,
          RecreateOnResume? recreateOnResume}) =>
      streamSubscription != null
          ? _streamSubscriptionsLifecycleAware
              .add(StreamSubscriptionLifecycleAware(
              streamSubscription: streamSubscription,
              lifecycleStates: lifecycleStates,
              recreateOnResume: recreateOnResume,
            ))
          : false;

  bool removeLifecycleAware(StreamSubscription? streamSubscription) {
    if (streamSubscription == null) return false;
    bool removed = false;
    _streamSubscriptionsLifecycleAware.removeWhere(
        (streamSubscriptionLifecycleAware) =>
            streamSubscriptionLifecycleAware.streamSubscription ==
            streamSubscription);
    return removed;
  }

  void _notifyLifecycleChange(LifecycleState lifecycleState) {
    if (lifecycleState != LifecycleState.resumed &&
        lifecycleState != LifecycleState.paused &&
        lifecycleState != LifecycleState.stopped &&
        lifecycleState != LifecycleState.destroyed) return;
    for (var streamSubscriptionLifecycleAware
        in _streamSubscriptionsLifecycleAware) {
      streamSubscriptionLifecycleAware.onLifecycleChanged(lifecycleState);
    }
  }

  @override
  void initState() {
    super.initState();
    _ambiguate(WidgetsBinding.instance)?.addObserver(this);
    _ambiguate(WidgetsBinding.instance)?.addPostFrameCallback((_) {
      _notifyLifecycleChange(_lifecycleState = LifecycleState.created);
      onCreated();
    });
  }

  @override
  void dispose() {
    _ambiguate(WidgetsBinding.instance)?.removeObserver(this);
    _lifecycleState = LifecycleState.destroyed;
    _notifyLifecycleChange(lifecycleState);
    onDestroy();
    super.dispose();
  }

  @override
  void deactivate() {
    if (_lifecycleState == LifecycleState.paused) {
      _lifecycleState = LifecycleState.resumed;
      _notifyLifecycleChange(lifecycleState);
      onResumed();
    } else {
      _lifecycleState = LifecycleState.paused;
      _notifyLifecycleChange(lifecycleState);
      onPaused();
    }
    super.deactivate();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    switch (state) {
      case AppLifecycleState.resumed:
        _lifecycleState = LifecycleState.resumed;
        _notifyLifecycleChange(lifecycleState);
        onResumed();
        break;
      case AppLifecycleState.inactive:
        _lifecycleState = LifecycleState.inactive;
        _notifyLifecycleChange(lifecycleState);
        onInactive();
        break;
      case AppLifecycleState
            .paused: //AppLifecycleState.paused is used as stopped state because deactivate() works more as a pause for lifecycle
        _lifecycleState = LifecycleState.stopped;
        _notifyLifecycleChange(lifecycleState);
        onStopped();
        break;
      case AppLifecycleState.detached:
        _lifecycleState = LifecycleState.detached;
        _notifyLifecycleChange(lifecycleState);
        onDetached();
        break;
      default:
        break;
    }
  }

  void refresh([VoidCallback? fn, bool forceRefresh = false]) {
    if (forceRefresh || mounted) setState(fn ?? () {});
  }

  @mustCallSuper
  @override
  Widget build(BuildContext context) {
    _isBuildCompleted = false;
    _ambiguate(WidgetsBinding.instance)
        ?.addPostFrameCallback((_) => onBuildCompleted());
    return const SizedBox();
  }

  bool get isLoading => _isLoading;

  Widget _buildProgressIndicator() =>
      _progressIndicator ??
      Column(
        mainAxisAlignment: MainAxisAlignment.center,
        mainAxisSize: MainAxisSize.min,
        children: <Widget>[
          _showProgressIndeterminate
              ? const CircularProgressIndicator()
              : const SizedBox(),
          const SizedBox(height: 16),
          if (_loadingMessage != null && _loadingMessage!.isNotEmpty)
            Text(
              _loadingMessage!,
              textAlign: TextAlign.center,
              style: const TextStyle(color: Colors.white),
            ),
        ],
      );

  Widget bodyLoading({
    required final Widget body,
    final bool? inAsyncCall,
    final double opacity = 0.7,
    final Color color = Colors.black,
    final Widget? progressIndicator,
    final Offset? offset,
    final bool dismissible = false,
  }) =>
      ModalProgressHUD(
        inAsyncCall: inAsyncCall ?? _isLoading,
        opacity: opacity,
        color: color,
        progressIndicator: progressIndicator ?? _buildProgressIndicator(),
        offset: offset,
        dismissible: dismissible,
        child: body,
      );

  WillPopScope bodyOnBackPressed({
    required final Widget body,
  }) =>
      WillPopScope(
        onWillPop: onBackPressed,
        child: body,
      );

  Widget setFragment(Widget fragment, {bool refresh = true}) {
    if (fragmentsManager.isNotEmpty) fragmentsManager.pop();
    return addFragment(fragment, refresh: refresh);
  }

  Widget addFragment(Widget fragment,
      {bool repeated = false, bool refresh = true}) {
    if (!repeated && fragmentsManager.isNotEmpty) {
      Widget currentFragment = fragmentsManager.peek();
      if (currentFragment.runtimeType == fragment.runtimeType &&
          currentFragment.key == fragment.key) return fragment;
    }
    this.fragment = fragmentsManager.push(fragment);
    if (refresh) setState(() {});
    return fragment;
  }

  void clearFragments() {
    fragment = const SizedBox();
    fragmentsManager.clear();
  }

  bool isCurrentFragment(Widget fragment) =>
      fragmentsManager.isNotEmpty ? fragmentsManager.peek() == fragment : false;

  ScaffoldFeatureController<SnackBar, SnackBarClosedReason> showSnackBar(
    SnackBar snackBar, {
    SnackBarClosedReason? hideCurrent,
    BuildContext? context,
  }) {
    context ??= this.context;
    if (hideCurrent != null) {
      ScaffoldMessenger.of(context).hideCurrentSnackBar(reason: hideCurrent);
    }
    return ScaffoldMessenger.of(context).showSnackBar(snackBar);
  }

  @override
  void showLoading(
      {String? message,
      bool showProgressIndeterminate = true,
      Widget? progressIndicator}) {
    _loadingMessage = message;
    _showProgressIndeterminate = showProgressIndeterminate;
    _progressIndicator = progressIndicator;
    refresh(() => _isLoading = true);
  }

  @override
  void hideLoading({int? delayMillis}) {
    if (delayMillis != null) {
      Future.delayed(Duration(milliseconds: delayMillis), () => hideLoading());
      return;
    }
    refresh(() => _isLoading = false);
  }

  @override
  void extras() {}

  @override
  void init() {}

  @override
  void check() {}

  @override
  void listeners() {}

  @override
  bool validate() => true;

  @override
  int getSelectedDrawerItemId() => 0;

  @override
  bool isPressBackAgain() => _pressBackAgain;

  @override
  void pressBackAgain(bool pressBackAgain,
      {String pressBackAgainMessage = 'Press back again to exit'}) {
    _pressBackAgain = pressBackAgain;
    _pressBackAgainMessage = pressBackAgainMessage;
  }

  @override
  void onBuildCompleted() {
    _isBuildCompleted = true;
  }

  showPressBackAgainMessage() => Fluttertoast.showToast(
      msg: _pressBackAgainMessage, toastLength: Toast.LENGTH_LONG);

  /// onBackPressed returns false to stop propagating the event, true in other cases
  @override
  Future<bool> onBackPressed() async {
    if (fragmentsManager.size > 1) {
      fragmentsManager.pop();
      fragment = fragmentsManager.peek();
      setState(() {});
      return false;
    }
    if (isPressBackAgain()) {
      if ((DateTime.now().millisecondsSinceEpoch - _exitControl) <= 3000) {
        return true;
      } else {
        _exitControl = DateTime.now().millisecondsSinceEpoch;
        showPressBackAgainMessage();
        return false;
      }
    }
    return true;
  }

  @override
  void pop<I extends Object?>({BuildContext? context, I? result}) {
    context ??= this.context;
    Navigator.pop(context, result);
  }

  @override
  Future<I?>? push<I>(
      {BuildContext? context,
      Route<I>? route,
      String? routeName,
      Widget? widget}) async {
    context ??= this.context;
    if (route != null) {
      return Navigator.push(context, route);
    } else if (widget != null) {
      return push(
          context: context,
          route: MaterialPageRoute(
              builder: (context) => widget,
              settings:
                  routeName != null ? RouteSettings(name: routeName) : null));
    }
    return null;
  }

  @override
  Future<I?>? pushReplacement<I, TO>(
      {BuildContext? context,
      Route<I>? newRoute,
      String? newRouteName,
      Widget? newWidget,
      TO? result}) async {
    context ??= this.context;
    if (newRoute != null) {
      return Navigator.pushReplacement(context, newRoute, result: result);
    } else if (newWidget != null) {
      return pushReplacement(
          context: context,
          newRoute: MaterialPageRoute(
              builder: (context) => newWidget,
              settings: newRouteName != null
                  ? RouteSettings(name: newRouteName)
                  : null),
          result: result);
    }
    return null;
  }

  @override
  void onClick({Widget? w, id, BuildContext? ctx}) {}

  @override
  void onLongClick({Widget? w, id, BuildContext? ctx}) {}

  @override
  void onItemLongClick(
      {Widget? w, T? item, int? position, id, BuildContext? ctx}) {}

  @override
  void onItemClick(
      {Widget? w, T? item, int? position, id, BuildContext? ctx}) {}

  @override
  void onCreated() {
//    print("onCreated");
  }

  @override
  void onResumed() {
//    print("onResumed");
  }

  @override
  void onInactive() {
//    print("onInactive");
  }

  @override
  void onPaused() {
//    print("onPaused");
  }

  @override
  void onDetached() {
//    print("onSuspending");
  }

  @override
  void onStopped() {
//    print("onStopped");
  }

  @override
  void onDestroy() {
//    print("onDestroy");
  }
}
