import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/extendables/adept_state.dart';
import 'package:adeptutils/src/exception_utils.dart';

///This wrapper is used to wrap some useful fields into a final object to be used in the whole StatefulWidget
class StatefulWrapper<S extends AdeptState, B> {
  S? state;
  B? base;
  BuildContext? context;
}

abstract class AdeptStatefulWidget<S extends AdeptState, T, B>
    extends StatefulWidget implements IWidget<T> {
  final _wrapper = StatefulWrapper<S, B>();

  set state(S? state) => _wrapper.state = state;
  S? get state => _wrapper.state;

  set _base(B? base) => _wrapper.base = base;
  B? get base => _wrapper.base;

  BuildContext? get context => _wrapper.context;

  AdeptStatefulWidget({Key? key, B? base})
      : super(
          key: key,
        ) {
    _base = base;
  }

  void handleException(dynamic e,
      {BuildContext? context,
      String title = 'Error',
      String btnContinueText = 'CONTINUE',
      String btnExitText = 'EXIT'}) {
    context ??= this.context;
    ExceptionUtils.handleException(e, context!,
        title: title,
        btnContinueText: btnContinueText,
        btnExitText: btnExitText);
  }

  @override
  void showLoading(
      {String? message,
      bool showProgressIndeterminate = true,
      Widget? progressIndicator}) {
    state?.showLoading(
        message: message,
        showProgressIndeterminate: showProgressIndeterminate,
        progressIndicator: progressIndicator);
  }

  @override
  void hideLoading({int? delayMillis}) {
    state?.hideLoading(delayMillis: delayMillis);
  }

  bool get isLoading => state?.isLoading ?? false;

  @override
  void extras() {
    state?.extras();
  }

  @override
  void init() {
    state?.init();
  }

  @override
  void check() {
    state?.check();
  }

  @override
  void listeners() {
    state?.listeners();
  }

  @override
  bool validate() => state?.validate() ?? true;

  @override
  int getSelectedDrawerItemId() => state?.getSelectedDrawerItemId() ?? 0;

  @override
  bool isPressBackAgain() => state?.isPressBackAgain() ?? false;

  @override
  void pressBackAgain(bool pressBackAgain,
      {String pressBackAgainMessage = 'Press back again to exit'}) {
    state?.pressBackAgain(pressBackAgain,
        pressBackAgainMessage: pressBackAgainMessage);
  }

  @override
  void onBuildCompleted() {
    state?.onBuildCompleted();
  }

  /// onBackPressed returns false to stop propagating the event, true in other cases
  @override
  Future<bool> onBackPressed() => state?.onBackPressed() ?? Future(() => true);

  @override
  void pop<I extends Object?>({BuildContext? context, I? result}) {
    context ??= this.context;
    Navigator.pop(context!, result);
  }

  @override
  Future<I?>? push<I>(
          {BuildContext? context,
          Route<I>? route,
          String? routeName,
          Widget? widget}) async =>
      state?.push(
          context: context, route: route, routeName: routeName, widget: widget);

  @override
  Future<I?>? pushReplacement<I, TO>(
          {BuildContext? context,
          Route<I>? newRoute,
          String? newRouteName,
          Widget? newWidget,
          TO? result}) async =>
      state?.pushReplacement(
          context: context,
          newRoute: newRoute,
          newRouteName: newRouteName,
          newWidget: newWidget,
          result: result);

  @override
  void onClick({Widget? w, id, BuildContext? ctx}) {
    state?.onClick(w: w, id: id);
  }

  @override
  void onLongClick({Widget? w, id, BuildContext? ctx}) {
    state?.onLongClick(w: w, id: id);
  }

  @override
  void onItemLongClick(
      {Widget? w, T? item, int? position, id, BuildContext? ctx}) {
    state?.onItemLongClick(w: w, item: item, position: position, id: id);
  }

  @override
  void onItemClick({Widget? w, T? item, int? position, id, BuildContext? ctx}) {
    state?.onItemClick(w: w, item: item, position: position, id: id);
  }

  @override
  void onCreated() {}

  @override
  void onResumed() {}

  @override
  void onInactive() {}

  @override
  void onPaused() {}

  @override
  void onDetached() {}

  @override
  void onStopped() {}

  @override
  void onDestroy() {}
}
