import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/exception_utils.dart';
import 'package:adeptutils/src/enumerators/lifecycle_state.dart';
import 'package:flutter/scheduler.dart';

/// This allows a value of type T or T? to be treated as a value of type T?.
///
/// We use this so that APIs that have become non-nullable can still be used
/// with `!` and `?` on the stable branch.
T? _ambiguate<T>(T? value) => value;

///This wrapper is used to wrap some useful fields into a final object to be used in the whole StatelessWidget
class StatelessWrapper<B> {
  B? base;
  late BuildContext context;
  LifecycleState? lifecycleState;
  Set<StreamSubscription> streamSubscriptions = {};
}

abstract class AdeptStatelessWidget<T, B> extends StatelessWidget
    with WidgetsBindingObserver
    implements IWidget<T> {
  final _wrapper = StatelessWrapper<B>();

  set _base(B? base) => _wrapper.base = base;
  B? get base => _wrapper.base;

  set _context(BuildContext context) => _wrapper.context = context;
  BuildContext get context => _wrapper.context;

  set _lifecycleState(LifecycleState lifecycleState) =>
      _wrapper.lifecycleState = lifecycleState;
  LifecycleState? get lifecycleState => _wrapper.lifecycleState;

  Set<StreamSubscription> get _streamSubscriptions =>
      _wrapper.streamSubscriptions;

  AdeptStatelessWidget({Key? key, B? base})
      : super(
          key: key,
        ) {
    _base = base;
  }

  ScaffoldFeatureController<SnackBar, SnackBarClosedReason>? showSnackBar(
    SnackBar snackBar, {
    SnackBarClosedReason? hideCurrent,
    BuildContext? context,
  }) {
    context ??= this.context;
    if (hideCurrent != null) {
      ScaffoldMessenger.of(context).hideCurrentSnackBar(reason: hideCurrent);
    }
    return ScaffoldMessenger.of(context).showSnackBar(snackBar);
  }

  void handleException(dynamic e,
      {BuildContext? context,
      String title = 'Error',
      String btnContinueText = 'CONTINUE',
      String btnExitText = 'EXIT'}) {
    context ??= this.context;
    ExceptionUtils.handleException(e, context,
        title: title,
        btnContinueText: btnContinueText,
        btnExitText: btnExitText);
  }

  bool addLifecycleAware(StreamSubscription? streamSubscription) =>
      streamSubscription != null
          ? _streamSubscriptions.add(streamSubscription)
          : false;

  bool removeLifecycleAware(StreamSubscription? streamSubscription) =>
      streamSubscription != null
          ? _streamSubscriptions.remove(streamSubscription)
          : false;

  void _notifyLifecycleChange(LifecycleState? lifecycleState) {
    if (lifecycleState == null) return;
    if (lifecycleState != LifecycleState.resumed &&
        lifecycleState != LifecycleState.paused &&
        lifecycleState != LifecycleState.stopped &&
        lifecycleState != LifecycleState.destroyed) return;
    for (var streamSubscription in _streamSubscriptions) {
      switch (lifecycleState) {
        case LifecycleState.resumed:
          streamSubscription.resume();
          break;
        case LifecycleState.paused:
        case LifecycleState.stopped:
          streamSubscription.pause();
          break;
        case LifecycleState.destroyed:
          streamSubscription.cancel();
          break;
        default:
      }
    }
  }

  @mustCallSuper
  @override
  Widget build(BuildContext context) {
    _context = context;
    _ambiguate(WidgetsBinding.instance)?.addObserver(this);
    _ambiguate(SchedulerBinding.instance)?.addPostFrameCallback((_) {
      _notifyLifecycleChange(_lifecycleState = LifecycleState.created);
      onCreated();
      onBuildCompleted();
    });
    return const SizedBox();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    switch (state) {
      case AppLifecycleState.resumed:
        _lifecycleState = LifecycleState.resumed;
        _notifyLifecycleChange(lifecycleState);
        onResumed();
        break;
      case AppLifecycleState.inactive:
        _lifecycleState = LifecycleState.inactive;
        _notifyLifecycleChange(lifecycleState);
        onInactive();
        break;
      case AppLifecycleState
            .paused: //AppLifecycleState.paused is used as stopped state because deactivate() works more as a pause for lifecycle
        _lifecycleState = LifecycleState.stopped;
        _notifyLifecycleChange(lifecycleState);
        onStopped();
        break;
      case AppLifecycleState.detached:
        _lifecycleState = LifecycleState.detached;
        _notifyLifecycleChange(lifecycleState);
        onDetached();
        break;
      default:
        break;
    }
  }

  WillPopScope bodyOnBackPressed({
    required final Widget body,
  }) =>
      WillPopScope(
        onWillPop: onBackPressed,
        child: body,
      );

  @override
  void showLoading(
      {String? message,
      bool showProgressIndeterminate = true,
      Widget? progressIndicator}) {}

  @override
  void hideLoading({int? delayMillis}) {}

  @override
  void extras() {}

  @override
  void init() {}

  @override
  void check() {}

  @override
  void listeners() {}

  @override
  bool validate() => true;

  @override
  int getSelectedDrawerItemId() => 0;

  @override
  bool isPressBackAgain() => false;

  @override
  void pressBackAgain(bool pressBackAgain,
      {String pressBackAgainMessage = 'Press back again to exit'}) {}

  @override
  void onBuildCompleted() {}

  @override
  Future<bool> onBackPressed() => Future(() => true);

  @override
  void pop<I extends Object?>({BuildContext? context, I? result}) {
    context ??= this.context;
    Navigator.pop(context, result);
  }

  @override
  Future<I?>? push<I>(
      {BuildContext? context,
      Route<I>? route,
      String? routeName,
      Widget? widget}) async {
    context ??= this.context;
    if (route != null) {
      return Navigator.push(context, route);
    } else if (widget != null) {
      return Navigator.push(
          context,
          MaterialPageRoute(
              builder: (context) => widget,
              settings:
                  routeName != null ? RouteSettings(name: routeName) : null));
    }
    return null;
  }

  @override
  Future<I?>? pushReplacement<I, TO>(
      {BuildContext? context,
      Route<I>? newRoute,
      String? newRouteName,
      Widget? newWidget,
      TO? result}) async {
    context ??= this.context;
    if (newRoute != null) {
      return Navigator.pushReplacement(context, newRoute);
    } else if (newWidget != null) {
      return Navigator.pushReplacement(
          context,
          MaterialPageRoute(
              builder: (context) => newWidget,
              settings: newRouteName != null
                  ? RouteSettings(name: newRouteName)
                  : null));
    }
    return null;
  }

  @override
  void onClick({Widget? w, id, BuildContext? ctx}) {}

  @override
  void onLongClick({Widget? w, id, BuildContext? ctx}) {}

  @override
  void onItemLongClick(
      {Widget? w, T? item, int? position, id, BuildContext? ctx}) {}

  @override
  void onItemClick(
      {Widget? w, T? item, int? position, id, BuildContext? ctx}) {}

  @override
  void onCreated() {}

  @override
  void onResumed() {}

  @override
  void onInactive() {}

  @override
  void onPaused() {}

  @override
  void onDetached() {}

  @override
  void onStopped() {}

  @override
  void onDestroy() {}
}
