import 'dart:async';
import 'dart:math' as Math;

import 'package:adeptutils/adeptutils.dart';
import 'package:adeptutils/src/interfaces/bool_callback.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:intl/intl.dart' as Intl;
import 'package:validators/validators.dart' as Validators;

class TextUtils {
  static bool isEmpty(String? text) =>
      text == null || text.isEmpty || text.trim().isEmpty;

  static bool isNotEmpty(String? text) => !isEmpty(text);

  static bool equals(String a, String b) => a == b;

  ///
  ///@param timeMillis the amount of time to convert in milliseconds
  ///@param decimals amount of decimals after float point when showing days, weeks, months and years
  ///@return
  ///
  static String toHumanReadableTimeCount(int timeMillis) {
    late String humanReadable = Duration(milliseconds: timeMillis).toString();
    try {
      humanReadable = humanReadable.split('.')[0];
    } catch (e) {
      print(e);
    }
    return humanReadable;

//    int integer;
//    double time = timeMillis?.toDouble();
//    if(time < TimeUtils.minute) {
//      return Intl.NumberFormat("00:00:#00", "en_US").format(time / TimeUtils.second);
//    }
//
//    if(time < TimeUtils.hour) {
//      int mins = (time / TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "00:"
//          "${Intl.NumberFormat("#00", "en_US").format(mins)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(seconds)}";
//    }
//    if(time < TimeUtils.day) {
//      int hours = (time / TimeUtils.hour).truncate();
//      int mins = ((time / TimeUtils.minute) % TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "${Intl.NumberFormat("#00", "en_US").format(hours)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(mins)}:"
//          "${Intl.NumberFormat("#00", "en_US").format(seconds)}";
//    }
//
//    if(time < TimeUtils.week) {
//      int days = (time / TimeUtils.hour).truncate();
//      int mins = ((time / TimeUtils.minute) % TimeUtils.minute).truncate();
//      int seconds = (time % TimeUtils.minute).truncate();
//      return "${Intl.NumberFormat("#0.00", "en_US").format(hours)}";
//    }
//    return "$time d";
  }

  ///
  ///@param bytes the amount of bytes to convert
  ///@param binarySystem true means to convert using Binary System (divide by 1024), false means
  ///          International System (divide by 1000)
  ///@param rightNotation true means to show KiB, MiB, GiB, etc if system is false. false means
  ///          to show kB, MB, GB, etc whatever value system is.
  ///@param decimals amount of decimals after float point
  ///@return
  ///
  static String toHumanReadableByteCount(int bytes,
      {bool binarySystem = true,
      bool rightNotation = false,
      int decimals = 2}) {
    int unit = binarySystem ? 1024 : 1000;
    if (bytes < unit) return '$bytes B';
    int exp = (Math.log(bytes) / Math.log(unit)).truncate();
    String pre = (binarySystem ? 'kMGTPEZY' : 'KMGTPEZY')[exp - 1] +
        (rightNotation ? (binarySystem ? 'i' : '') : '');
    if (!rightNotation) pre = pre.toUpperCase();
    String decimalsParam = '';
    if (decimals > 0) decimalsParam = '0.';
    while (decimals-- > 0) {
      decimalsParam += '0';
    }
    return Intl.NumberFormat('#$decimalsParam ${pre}B', 'en_US')
        .format(bytes / Math.pow(unit, exp));
  }

  ///
  ///@param distance: distance in meters
  ///@param advanced: true indicates distance from millimeters to kilometers, otherwise indicates distance in meters or kilometers depending on the distance; only for metric system
  ///@param metric: indicates the unit system: metric or imperial system
  ///@param locale: indicates the user locale to get the proper unit system: metric or imperial system
  ///@param context: indicates the build context to be used to get the system locale. This will be used if locale param is null
  ///@param decimals: amount of decimals after float point
  ///@return formatted: distance as String
  ///
  static String toHumanReadableDistanceCount(double distance,
      {bool advanced = false,
      bool? metric,
      Locale? locale,
      BuildContext? context,
      int decimals = 2}) {
    metric ??= (locale ??=
            FlutterViewUtils.get(context: context).platformDispatcher.locale)
        .usesMetricSystem;
    String unit;
    String decimalsParam = '';
    if (decimals > 0) decimalsParam = '0.';
    while (decimals-- > 0) {
      decimalsParam += '0';
    }
    if (metric) {
      if (advanced) {
        int min = 10;
        if (distance < min) return '$distance mm';
        double log10(num x) {
          num log = Math.log(x) / Math.log(10) + 0.000001;
          if ((log - log.truncate()) < 0.0001) log = log.truncateToDouble();
          return log as double;
        }

        int exp = (log10(distance) / log10(min)).truncate();
        unit = ['cm', 'dm', 'm', 'dam', 'hm', 'km'][exp - 1];
        distance /= Math.pow(min, exp);
      } else {
        if (distance < 1000) {
          unit = 'm';
        } else {
          distance /= 1000;
          unit = 'km';
        }
      }
    } else {
      if (distance < 1609.344) {
        distance *= 3.2808399;
        unit = 'ft';
      } else {
        distance /= 1609.344;
        unit = 'mi';
      }
    }
    return Intl.NumberFormat('#$decimalsParam $unit', locale.toString())
        .format(distance);
  }

  /// The formula is speed m/time where the speed should be in meters and time is specified in seconds base
  /// for instance:
  ///   to indicate 1000 m/seconds the call should be (1000, 1)
  ///   to indicate 1000 m/minute the call should be (1000, 60)
  ///   to indicate 1000 m/hour the call should be (1000, 3600)
  /// @param speed: in meters
  /// @param time: in seconds has to be 1s or 60s or 3600s; that indicates seconds, minute, hour
  ///@param metric: indicates the unit system: metric or imperial system
  /// @param locale: indicates the user locale to get the proper unit system: metric or imperial system
  ///@param context: indicates the build context to be used to get the system locale. This will be used if locale param is null
  /// @param decimals: amount of decimals after float point
  /// @return formatted: speed in km/h or mi/h according to metric as String
  ///
  static String toHumanReadableSpeedCount(double speed, int time,
      {bool? metric, Locale? locale, BuildContext? context, int decimals = 2}) {
    metric ??= (locale ??=
            FlutterViewUtils.get(context: context).platformDispatcher.locale)
        .usesMetricSystem;
    if (time == 1) {
      time = 3600;
    } else if (time == 60) {
      time = 60;
    } else if (time == 3600) {
      time = 1;
    } else {
      time = 1;
    }

    speed = speed * time.toDouble() / (metric ? 1000 : 1609.344).toDouble();
    String decimalsParam = '';
    if (decimals > 0) decimalsParam = '0.';
    while (decimals-- > 0) {
      decimalsParam += '0';
    }
    String unit = metric ? 'km' : 'mi';
    return Intl.NumberFormat('#$decimalsParam $unit/h', locale.toString())
        .format(speed);
  }
}

typedef Validation = bool Function({AdeptController? controller});

class ValidationController {
  final Validation? isValid;
  final String? message;

  ValidationController({this.isValid, this.message});
}

class EmptyValidationController extends ValidationController {
  EmptyValidationController({String message = 'This field is required'})
      : super(
          message: message,
          isValid: ({controller}) =>
              TextUtils.isNotEmpty(controller?.rawValue?.toString()),
        );
}

class LengthValidationController extends ValidationController {
  LengthValidationController(
      {int minLength = 0,
      int? maxLength,
      String message = 'Invalid length',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isLength(textValue!, minLength, maxLength);
            });
}

class RepeatPasswordValidationController extends ValidationController {
  RepeatPasswordValidationController(
      {AdeptController? passwordController,
      String message = "Passwords doesn't match"})
      : super(
          message: message,
          isValid: ({controller}) =>
              controller?.rawValue?.toString() ==
              passwordController?.rawValue?.toString(),
        );
}

class EmailValidationController extends ValidationController {
  EmailValidationController(
      {String message = 'Invalid email address', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isEmail(textValue!);
            });
}

class UrlValidationController extends ValidationController {
  UrlValidationController(
      {String message = 'Invalid url', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isURL(textValue);
            });
}

class IPValidationController extends ValidationController {
  IPValidationController(
      {String message = 'Invalid ip address', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isIP(textValue);
            });
}

class AlphanumericValidationController extends ValidationController {
  AlphanumericValidationController(
      {String message = 'Invalid alphanumeric', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isAlphanumeric(textValue!);
            });
}

class CreditCardValidationController extends ValidationController {
  CreditCardValidationController(
      {String message = 'Invalid credit card', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isCreditCard(textValue!);
            });
}

class DateValidationController extends ValidationController {
  DateValidationController(
      {String message = 'Invalid date', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isDate(textValue!);
            });
}

class AsciiValidationController extends ValidationController {
  AsciiValidationController(
      {String message = 'Invalid Ascii', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isAscii(textValue!);
            });
}

class Base64ValidationController extends ValidationController {
  Base64ValidationController(
      {String message = 'Invalid Base64', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isBase64(textValue!);
            });
}

class OnlyTextValidationController extends ValidationController {
  OnlyTextValidationController(
      {String message = 'Text should contain only text', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isAlpha(textValue!);
            });
}

class DNSValidationController extends ValidationController {
  DNSValidationController(
      {bool requireTld = true,
      bool allowUnderscores = false,
      String message = 'Invalid DNS',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isFQDN(textValue!);
            });
}

class HexadecimalValidationController extends ValidationController {
  HexadecimalValidationController(
      {String message = 'Invalid hexadecimal', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isHexadecimal(textValue!);
            });
}

class HexColorValidationController extends ValidationController {
  HexColorValidationController(
      {String message = 'Invalid Hexcolor', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isHexColor(textValue!);
            });
}

class ISBNValidationController extends ValidationController {
  ISBNValidationController(
      {dynamic version, String message = 'Invalid ISBN', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isISBN(textValue, version);
            });
}

class JsonValidationController extends ValidationController {
  JsonValidationController(
      {String message = 'Invalid Json', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isJSON(textValue);
            });
}

class LowercaseValidationController extends ValidationController {
  LowercaseValidationController(
      {String message = 'Invalid lowercase', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isLowercase(textValue!);
            });
}

class UppercaseValidationController extends ValidationController {
  UppercaseValidationController(
      {String message = 'Invalid Uppercase', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isUppercase(textValue!);
            });
}

class MongoIdValidationController extends ValidationController {
  MongoIdValidationController(
      {String message = 'Invalid MongoId', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isMongoId(textValue!);
            });
}

class SurrogatePairValidationController extends ValidationController {
  SurrogatePairValidationController(
      {String message = 'Invalid SurrogatePair', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isSurrogatePair(textValue!);
            });
}

class UUIDValidationController extends ValidationController {
  UUIDValidationController(
      {dynamic version, String message = 'Invalid UUID', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return Validators.isUUID(textValue, version);
            });
}

class NumericValidationController extends ValidationController {
  NumericValidationController(
      {String message = 'Text must be a number', bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return num.tryParse(textValue!) != null;
            });
}

class IntegerValidationController extends ValidationController {
  IntegerValidationController(
      {String message = 'Text must be an integer number',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return int.tryParse(textValue!) != null;
            });
}

class DecimalValidationController extends ValidationController {
  DecimalValidationController(
      {String message = 'Text must be an decimal number',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              return double.tryParse(textValue!) != null;
            });
}

class PositiveIntegerValidationController extends ValidationController {
  PositiveIntegerValidationController(
      {String message = 'Text must be an integer positive number',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              int? value = int.tryParse(textValue!);
              return (value ?? -1) >= 0;
            });
}

class PositiveDecimalValidationController extends ValidationController {
  PositiveDecimalValidationController(
      {String message = 'Text must be a decimal positive number',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              num? value = num.tryParse(textValue!);
              return (value ?? -1) >= 0;
            });
}

class IntegerRangeValidationController extends ValidationController {
  IntegerRangeValidationController(
      {int? minValue,
      int? maxValue,
      String message = 'Number must fall in the specified range',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              int? value = int.tryParse(textValue!);
              return (value ?? 0) >= (minValue ?? 0) &&
                  (value ?? 0) <= (maxValue ?? 0);
            });
}

class DecimalRangeValidationController extends ValidationController {
  DecimalRangeValidationController(
      {num? minValue,
      num? maxValue,
      String message = 'Number must fall in the specified range',
      bool required = false})
      : super(
            message: message,
            isValid: ({controller}) {
              String? textValue = controller?.rawValue?.toString();
              if (!required && TextUtils.isEmpty(textValue)) return true;
              num? value = num.tryParse(textValue!);
              return (value ?? 0) >= (minValue ?? 0) &&
                  (value ?? 0) <= (maxValue ?? 0);
            });
}

mixin AdeptController<T> {
  List<ValidationController> validations = [];
  BoolCallback? validationsDependency;
  FocusNode? focusNode;
  String? error;

  bool get hasValidations => validations.isNotEmpty;

  bool get isValidationsDependencySatisfied =>
      validationsDependency?.call() ?? true;

  bool get canValidate => hasValidations && isValidationsDependencySatisfied;

  bool validate({bool notify = true}) {
    clearError(notify: notify);
    if (!canValidate) return true;
    for (ValidationController validation in validations) {
      if (!(validation.isValid?.call(controller: this) ?? true)) {
        setError(validation.message, notify: notify);
        return false;
      }
    }
    return true;
  }

  bool get isValid {
    if (!canValidate) return true;
    for (ValidationController validation in validations) {
      if (!(validation.isValid?.call(controller: this) ?? true)) return false;
    }
    return true;
  }

  String? get validationMessage {
    if (!canValidate) return null;
    for (ValidationController validation in validations) {
      if (!(validation.isValid?.call(controller: this) ?? true)) {
        return validation.message;
      }
    }
    return null;
  }

  T get rawValue;

  dynamic get controllerValue => rawValue;

  bool get hasError => error?.isNotEmpty ?? false;

  bool get isEmpty;

  bool get isNotEmpty => !isEmpty;

  void setError(String? error, {bool notify = true});

  void clearError({bool notify = true});

  void clear();

  void _feedback() async {
    for (int i = 0; i < 3; ++i) {
      await Future.delayed(
          const Duration(milliseconds: 100), () => HapticFeedback.vibrate());
    }
  }
}

class AdeptTextEditingController extends TextEditingController
    with AdeptController<String> {
  AdeptTextEditingController({
    String? text,
    FocusNode? focusNode,
    List<ValidationController>? validations,
    BoolCallback? validationsDependency,
  }) : super(text: text) {
    this.focusNode = focusNode;
    if (validations != null) this.validations = validations;
    this.validationsDependency = validationsDependency;
  }

  @override
  String get rawValue => text;

  @override
  get controllerValue => value;

  @override
  bool get isEmpty => text.isEmpty;

  @override
  void setError(String? error, {bool notify = true}) {
    this.error = error;
    if (notify) {
      if (hasError) _feedback();
      notifyListeners();
    }
  }

  @override
  void clearError({bool notify = false}) => setError(null, notify: notify);

  @override
  void clear() {
    clearError();
    super.clear();
  }
}

class AdeptValueController<T> extends ValueNotifier<T?>
    with AdeptController<T?> {
  AdeptValueController({
    T? value,
    FocusNode? focusNode,
    List<ValidationController>? validations,
    BoolCallback? validationsDependency,
  }) : super(value) {
    this.focusNode = focusNode;
    if (validations != null) this.validations = validations;
    this.validationsDependency = validationsDependency;
  }

  @override
  T? get rawValue => value;

  @override
  bool get isEmpty => value == null;

  @override
  void setError(String? error, {bool notify = true}) {
    this.error = error;
    if (notify) {
      if (hasError) _feedback();
      notifyListeners();
    }
  }

  @override
  void clearError({bool notify = false}) => setError(null, notify: notify);

  @override
  void clear() {
    clearError();
    value = null;
  }
}

class AdeptValueNSController<T> extends ValueNotifier<T>
    with AdeptController<T> {
  AdeptValueNSController({
    required T value,
    FocusNode? focusNode,
    List<ValidationController>? validations,
    BoolCallback? validationsDependency,
  }) : super(value) {
    this.focusNode = focusNode;
    if (validations != null) this.validations = validations;
    this.validationsDependency = validationsDependency;
  }

  @override
  T get rawValue => value;

  @override
  bool get isEmpty => value == null;

  @override
  void setError(String? error, {bool notify = true}) {
    this.error = error;
    if (notify) {
      if (hasError) _feedback();
      notifyListeners();
    }
  }

  @override
  void clearError({bool notify = false}) => setError(null, notify: notify);

  @override
  void clear() {
    clearError();
  }
}
