import 'package:adeptutils/adeptutils.dart';
import 'dart:math' as Math;

typedef OnLocationChanged = Function(Location location);
typedef OnStatusChanged = Function(String provider, int status, dynamic extras);
typedef OnProviderEnabled = Function(String provider);
typedef OnProviderDisabled = Function(String provider);

abstract class LocationListener {
  /// Called when the location has changed.
  /// <p> There are no restrictions on the use of the supplied Location object.
  /// @param location The new location, as a Location object.
  void onLocationChanged(Location location);

  ///
  /// Called when the provider status changes. This method is called when
  /// a provider is unable to fetch a location or if the provider has recently
  /// become available after a period of unavailability.
  ///
  /// @param provider the name of the location provider associated with this
  /// update.
  /// @param status {@link LocationProvider#OUT_OF_SERVICE} if the
  /// provider is out of service, and this is not expected to change in the
  /// near future; {@link LocationProvider#TEMPORARILY_UNAVAILABLE} if
  /// the provider is temporarily unavailable but is expected to be available
  /// shortly; and {@link LocationProvider#AVAILABLE} if the
  /// provider is currently available.
  /// @param extras an optional Bundle which will contain provider specific
  /// status variables.
  ///
  /// <p> A number of common key/value pairs for the extras Bundle are listed
  /// below. Providers that use any of the keys on this list must
  /// provide the corresponding value as described below.
  ///
  /// <ul>
  /// <li> satellites - the number of satellites used to derive the fix
  /// </ul>
  ///
  void onStatusChanged(String provider, int status, dynamic extras);

  ///
  /// Called when the provider is enabled by the user.
  ///
  /// @param provider the name of the location provider associated with this
  /// update.
  ///
  void onProviderEnabled(String provider);

  ///
  /// Called when the provider is disabled by the user. If requestLocationUpdates
  /// is called on an already disabled provider, this method is called
  /// immediately.
  ///
  /// @param provider the name of the location provider associated with this
  /// update.
  ///
  void onProviderDisabled(String provider);
}

class Location extends Jsonable<Location> {
  /// The latitude of this location in degrees normalized to the interval -90.0 to +90.0 (both inclusive).
  final double latitude;

  /// The longitude of the location in degrees normalized to the interval -180 (exclusive) to +180 (inclusive).
  final double longitude;

  /// The time at which this location was determined.
  final DateTime timestamp;

  ///Indicate if location was created from a mock provider.
  ///
  /// The mock information is not available on all devices. In these cases the returned value is false.
  final bool mocked;

  /// The altitude of the device in meters.
  ///
  /// The altitude is not available on all devices. In these cases the returned value is 0.0.
  final double altitude;

  /// The estimated horizontal accuracy of the location in meters.
  ///
  /// The accuracy is not available on all devices. In these cases the value is 0.0.
  final double accuracy;

  /// The heading in which the device is traveling in degrees.
  ///
  /// The heading is not available on all devices. In these cases the value is 0.0.
  final double heading;

  /// The speed at which the devices is traveling in meters per second over ground.
  ///
  /// The speed is not available on all devices. In these cases the value is 0.0.
  final double speed;

  /// The estimated speed accuracy of this location, in meters per second.
  ///
  /// The speedAccuracy is not available on all devices. In these cases the value is 0.0.
  final double speedAccuracy;

  Location({
    double? longitude,
    double? latitude,
    DateTime? timestamp,
    bool? mocked,
    double? accuracy,
    double? altitude,
    double? heading,
    double? speed,
    double? speedAccuracy,
  })  : longitude = longitude ?? 0.0,
        latitude = latitude ?? 0.0,
        timestamp = timestamp ?? DateTime.now(),
        mocked = mocked ?? false,
        accuracy = accuracy ?? 0.0,
        altitude = altitude ?? 0.0,
        heading = heading ?? 0.0,
        speed = speed ?? 0.0,
        speedAccuracy = speedAccuracy ?? 0.0;

  bool get isNotEmpty => !isEmpty;

  bool get isEmpty => (longitude == 0 && latitude == 0);

  bool isDifferentThan(Location? location) =>
      location == null ||
      (latitude != location.latitude && longitude != location.longitude);

  ///
  /// Using Haversine formula
  ///
  double distanceTo({double? latitude, double? longitude, Location? location}) {
    assert((latitude != null && longitude != null) || location != null);
    if (location != null) {
      return distanceTo(
          latitude: location.latitude, longitude: location.longitude);
    }
    return distanceBetween(
        this.latitude, this.longitude, latitude!, longitude!);
  }

  ///
  /// Using Euclidean formula
  ///
  double fastDistanceTo(
      {double? latitude, double? longitude, Location? location}) {
    assert((latitude != null && longitude != null) || location != null);
    if (location != null) {
      return fastDistanceTo(
          latitude: location.latitude, longitude: location.longitude);
    }
    return fastDistanceBetween(
        this.latitude, this.longitude, latitude!, longitude!);
  }

  Location copyWith({
    double? latitude,
    double? longitude,
    double? accuracy,
    double? altitude,
    double? heading,
    double? speed,
    double? speedAccuracy,
    DateTime? timestamp,
    bool? mocked,
  }) =>
      Location(
        latitude: latitude ?? this.latitude,
        longitude: longitude ?? this.longitude,
        accuracy: accuracy ?? this.accuracy,
        altitude: altitude ?? this.altitude,
        heading: heading ?? this.heading,
        speed: speed ?? this.speed,
        speedAccuracy: speedAccuracy ?? this.speedAccuracy,
        timestamp: timestamp ?? this.timestamp,
        mocked: mocked ?? this.mocked,
      );

  @override
  Map<String, dynamic> toJson() => {
        'longitude': longitude,
        'latitude': latitude,
        'timestamp': timestamp.toIso8601String(),
        'mocked': mocked,
        'accuracy': accuracy,
        'altitude': altitude,
        'heading': heading,
        'speed': speed,
        'speedAccuracy': speedAccuracy,
      };

  @override
  Location? fromJsonMap(Map<String, dynamic>? json) =>
      json == null ? null : Location.fromJson(json);

  factory Location.fromJson(Map<String, dynamic> json) => Location(
        longitude: (json['longitude'] as num?)?.toDouble(),
        latitude: (json['latitude'] as num?)?.toDouble(),
        timestamp: (json['timestamp'] as String?) == null
            ? null
            : DateTime.tryParse(json['timestamp']),
        mocked: json['mocked'] as bool?,
        accuracy: (json['accuracy'] as num?)?.toDouble(),
        altitude: (json['altitude'] as num?)?.toDouble(),
        heading: (json['heading'] as num?)?.toDouble(),
        speed: (json['speed'] as num?)?.toDouble(),
        speedAccuracy: (json['speedAccuracy'] as num?)?.toDouble(),
      );

  /// Returns the distance between 2 points of coordinates with Haversine formula
  ///
  /// @see https://en.wikipedia.org/wiki/Haversine_formula
  /// @see https://stackoverflow.com/a/1502821/4241030
  /// @param lat1 Latitude of the point A
  /// @param lng1 Longitude of the point A
  /// @param lat2 Latitude of the point B
  /// @param lng2 Longitude of the point B
  static double distanceBetween(
      double lat1, double lng1, double lat2, double lng2) {
    // The radius of the planet earth in meters
    const R = 6378137;
    final dLat = _degreesToRadians(lat2 - lat1);
    final dLong = _degreesToRadians(lng2 - lng1);
    final a = Math.sin(dLat / 2) * Math.sin(dLat / 2) +
        Math.cos(_degreesToRadians(lat1)) *
            Math.cos(_degreesToRadians(lat1)) *
            Math.sin(dLong / 2) *
            Math.sin(dLong / 2);

    final c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
    final distance = R * c;

    return distance;
  }

  /// Converts degrees to radians.
  ///
  /// @param degrees Number of degrees.
  static double _degreesToRadians(degrees) {
    return degrees * Math.pi / 180;
  }

  ///
  /// Calculate the Euclidean distance from this point to another using the Pythagorean theorem.
  ///
  /// @param other The point to calculate the distance to
  /// @return the distance in degrees as a double
  ///
  static double fastDistanceBetween(
      double lat1, double lng1, double lat2, double lng2) {
    return _hypot(lng1 - lng2, lat1 - lat2);
  }

  static double _hypot(double x, double y) {
    var first = x.abs();
    double second = y.abs();

    if (y > x) {
      first = y.abs();
      second = x.abs();
    }

    if (first == 0.0) {
      return second;
    }

    final t = second / first;
    return first * Math.sqrt(1 + t * t);
  }
}
